package sun.nio.ch;

import java.lang.ref.WeakReference;
import java.util.Iterator;
import java.nio.channels.OverlappingFileLockException;
import java.util.ArrayList;
import java.io.IOException;
import java.io.FileDescriptor;
import java.nio.channels.Channel;
import java.nio.channels.FileLock;
import java.lang.ref.ReferenceQueue;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;

class SharedFileLockTable extends FileLockTable
{
    private static ConcurrentHashMap<FileKey, List<SharedFileLockTable.FileLockReference>> lockMap;
    private static ReferenceQueue<FileLock> queue;
    private final Channel channel;
    private final FileKey fileKey;
    
    SharedFileLockTable(Channel channel, FileDescriptor fileDescriptor) throws IOException {
        this.channel = channel;
        this.fileKey = FileKey.create(fileDescriptor);
    }
    
    @Override
    public void add(FileLock fileLock) throws OverlappingFileLockException {
        List<SharedFileLockTable.FileLockReference> list = (List)SharedFileLockTable.lockMap.get(this.fileKey);
        while (true) {
            if (list == null) {
                ArrayList value = new ArrayList(2);
                synchronized (value) {
                    List<SharedFileLockTable.FileLockReference> list2 = (List<SharedFileLockTable.FileLockReference>)SharedFileLockTable.lockMap.putIfAbsent(this.fileKey, value);
                    if (list == null) {
                        value.add(new SharedFileLockTable.FileLockReference(fileLock, SharedFileLockTable.queue, this.fileKey));
                        break;
                    }
                }
            }
            synchronized (list) {
                List<Object> list3 = (List)SharedFileLockTable.lockMap.get(this.fileKey);
                if (list == list3) {
                    this.checkList(list, fileLock.position(), fileLock.size());
                    list.add(new SharedFileLockTable.FileLockReference(fileLock, SharedFileLockTable.queue, this.fileKey));
                    break;
                }
                list = list3;
            }
        }
        this.removeStaleEntries();
    }
    
    private void removeKeyIfEmpty(FileKey fileKey, List<SharedFileLockTable.FileLockReference> list) {
        assert !(!Thread.holdsLock(list));
        assert SharedFileLockTable.lockMap.get(fileKey) == list;
        if (list.isEmpty()) {
            SharedFileLockTable.lockMap.remove(fileKey);
        }
        return;
    }
    
    @Override
    public void remove(FileLock fileLock) {
        assert fileLock != null;
        List<Object> list = (List)SharedFileLockTable.lockMap.get(this.fileKey);
        if (list == null) {
            return;
        }
        synchronized (list) {
            int i = 0;
            while (i < list.size()) {
                SharedFileLockTable.FileLockReference fileLockReference = (SharedFileLockTable.FileLockReference)list.get(i);
                FileLock fileLock2 = (FileLock)fileLockReference.get();
                if (fileLock2 == fileLock) {
                    assert fileLock2 != null && fileLock2.acquiredBy() == this.channel;
                    fileLockReference.clear();
                    list.remove(i);
                    break;
                }
                else {
                    ++i;
                }
            }
        }
        return;
    }
    
    @Override
    public List<FileLock> removeAll() {
        ArrayList<FileLock> list = new ArrayList();
        List<SharedFileLockTable.FileLockReference> list2 = (List)SharedFileLockTable.lockMap.get(this.fileKey);
        if (list2 != null) {
            synchronized (list2) {
                int i = 0;
                while (i < list2.size()) {
                    SharedFileLockTable.FileLockReference fileLockReference = (SharedFileLockTable.FileLockReference)list2.get(i);
                    FileLock fileLock = (FileLock)fileLockReference.get();
                    if (fileLock != null && fileLock.acquiredBy() == this.channel) {
                        fileLockReference.clear();
                        list2.remove(i);
                        list.add(fileLock);
                    }
                    else {
                        ++i;
                    }
                }
                this.removeKeyIfEmpty(this.fileKey, list2);
            }
        }
        return list;
    }
    
    @Override
    public void replace(FileLock fileLock, FileLock fileLock2) {
        List list = (List)SharedFileLockTable.lockMap.get(this.fileKey);
        assert list != null;
        synchronized (list) {
            int i = 0;
            while (i < list.size()) {
                SharedFileLockTable.FileLockReference fileLockReference = (SharedFileLockTable.FileLockReference)list.get(i);
                if ((FileLock)fileLockReference.get() == fileLock) {
                    fileLockReference.clear();
                    list.set(i, new SharedFileLockTable.FileLockReference(fileLock2, SharedFileLockTable.queue, this.fileKey));
                    break;
                }
                ++i;
            }
        }
        return;
    }
    
    private void checkList(List<SharedFileLockTable.FileLockReference> list, long position, long size) throws OverlappingFileLockException {
        assert !(!Thread.holdsLock(list));
        Iterator<Object> iterator = list.iterator();
        while (iterator.hasNext()) {
            FileLock fileLock = (FileLock)((SharedFileLockTable.FileLockReference)iterator.next()).get();
            if (fileLock != null && fileLock.overlaps(position, size)) {
                throw new OverlappingFileLockException();
            }
            continue;
        }
        return;
    }
    
    private void removeStaleEntries() {
        SharedFileLockTable.FileLockReference fileLockReference;
        while ((fileLockReference = (SharedFileLockTable.FileLockReference)SharedFileLockTable.queue.poll()) != null) {
            FileKey fileKey = fileLockReference.fileKey();
            List<SharedFileLockTable.FileLockReference> list = (List)SharedFileLockTable.lockMap.get(fileKey);
            if (list != null) {
                synchronized (list) {
                    list.remove(fileLockReference);
                    this.removeKeyIfEmpty(fileKey, list);
                }
            }
            else {
                continue;
            }
        }
    }
    
    static {
        SharedFileLockTable.lockMap = new ConcurrentHashMap();
        SharedFileLockTable.queue = new ReferenceQueue();
    }
    
    private static class FileLockReference extends WeakReference<FileLock>
    {
        private FileKey fileKey;
        
        FileLockReference(FileLock referent, ReferenceQueue<FileLock> q, FileKey fileKey) {
            super(referent, q);
            this.fileKey = fileKey;
        }
        
        FileKey fileKey() {
            return this.fileKey;
        }
    }
}
